/*
 * Copyright 2020 LINE Corporation
 *
 * LINE Corporation licenses this file to you under the Apache License,
 * version 2.0 (the "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at:
 *
 *   https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

package com.linecorp.armeria.spring;

import static org.assertj.core.api.Assertions.assertThat;

import java.util.concurrent.atomic.AtomicReference;

import javax.annotation.Nullable;
import javax.inject.Inject;

import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.boot.autoconfigure.SpringBootApplication;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.context.annotation.Bean;
import org.springframework.stereotype.Component;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.context.junit4.SpringRunner;
import org.springframework.validation.annotation.Validated;

import com.linecorp.armeria.client.WebClient;
import com.linecorp.armeria.common.HttpResponse;
import com.linecorp.armeria.common.HttpStatus;
import com.linecorp.armeria.common.logging.RequestLog;
import com.linecorp.armeria.common.logging.RequestLogAccess;
import com.linecorp.armeria.server.Server;
import com.linecorp.armeria.server.annotation.Get;
import com.linecorp.armeria.spring.CgLibServiceNameTest.TestConfiguration;
import com.linecorp.armeria.spring.CgLibServiceNameTest.TestConfiguration.MyAnnotatedService;

@RunWith(SpringRunner.class)
@SpringBootTest(classes = TestConfiguration.class)
@ActiveProfiles({ "local", "autoConfTest" })
public class CgLibServiceNameTest {

    private static final AtomicReference<RequestLogAccess> logRef = new AtomicReference<>();

    @Nullable
    private static MyAnnotatedService proxiedMyAnnotatedService;

    @SpringBootApplication
    public static class TestConfiguration {

        @Validated
        @Component
        public class MyAnnotatedService {
            @Get("/hello")
            public HttpResponse world() {
                return HttpResponse.of(HttpStatus.OK);
            }
        }

        @Bean
        public ArmeriaServerConfigurator annotatedService(MyAnnotatedService annotatedService) {
            proxiedMyAnnotatedService = annotatedService;
            return sb -> {
                sb.annotatedService(annotatedService);
                sb.decorator((delegate, ctx, req) -> {
                    logRef.set(ctx.log());
                    return delegate.serve(ctx, req);
                });
            };
        }
    }

    @Inject
    private Server server;

    /**
     * Verifies that a default service name returned by RequestLog is a user-defined class name.
     * The user-defined class could be changed at runtime by Spring which uses cglib for generation of
     * dynamic proxies. For example, if a class is annotated with `@Validated`,
     * a proxy class of MyAnnotatedService, such as MyAnnotatedService$$EnhancerBySpringCGLIB$$1a2b3c4d,
     * will be injected instead.
     */
    @Test
    public void normalizedServiceName() {
        final WebClient client = WebClient.of("http://127.0.0.1:" + server.activeLocalPort() + '/');
        client.get("/hello").aggregate().join();
        final RequestLog requestLog = logRef.get().whenComplete().join();
        // Make sure that the injected MyAnnotatedService is generated byte code by cglib.
        assertThat(proxiedMyAnnotatedService.getClass().getSimpleName())
                .contains("MyAnnotatedService$$EnhancerBySpringCGLIB$$");
        assertThat(requestLog.serviceName()).isEqualTo(MyAnnotatedService.class.getName());
    }
}
